@@ -38,13 +38,14 @@ CA 94129, USA, for further information.
 import os
 import string
 from binascii import b2a_base64
+from collections import defaultdict
+from dataclasses import dataclass
+
 import pymupdf
 from pymupdf import mupdf
 from pymupdf4llm.helpers.get_text_lines import get_raw_lines, is_white
 from pymupdf4llm.helpers.multi_column import column_boxes
 from pymupdf4llm.helpers.progress import ProgressBar
-from dataclasses import dataclass
-from collections import defaultdict
 
 pymupdf.TOOLS.unset_quad_corrections(True)
 
@@ -316,6 +317,7 @@ def to_markdown(
     filename=None,
     force_text=True,
     page_chunks=False,
+    page_separators=False,
     margins=0,
     dpi=150,
     page_width=612,
@@ -341,6 +343,7 @@ def to_markdown(
         image_format: (str) use this image format. Choose a supported one.
         force_text: (bool) output text despite of image background.
         page_chunks: (bool) whether to segment output by page.
+        page_separators: (bool) whether to include page separators in output.
         margins: omit content overlapping margin areas.
         dpi: (int) desired resolution for generated images.
         page_width: (float) assumption if page layout is variable.
@@ -381,7 +384,7 @@ def to_markdown(
     IGNORE_IMAGES = ignore_images
     IGNORE_GRAPHICS = ignore_graphics
     DETECT_BG_COLOR = detect_bg_color
-    if doc.is_form_pdf or doc.has_annots():
+    if doc.is_form_pdf or (doc.is_pdf and doc.has_annots()):
         doc.bake()
 
     # for reflowable documents allow making 1 page for the whole document
@@ -560,6 +563,7 @@ def to_markdown(
                         )
                         parms.line_rects.extend(cells)
                     parms.written_tables.append(i)
+                    prev_hdr_string = None
 
             # ------------------------------------------------------------
             # Pick up images / graphics ABOVE this text block
@@ -569,7 +573,7 @@ def to_markdown(
                     if i in parms.written_images:
                         continue
                     r = parms.img_rects[i]
-                    if r.y1 <= lrect.y0 and (
+                    if max(r.y0, lrect.y0) < min(r.y1, lrect.y1) and (
                         0
                         or lrect.x0 <= r.x0 < lrect.x1
                         or lrect.x0 < r.x1 <= lrect.x1
@@ -592,6 +596,7 @@ def to_markdown(
                             if not is_white(img_txt):
                                 out_string += img_txt
                         parms.written_images.append(i)
+                        prev_hdr_string = None
 
             parms.line_rects.append(lrect)
             # if line rect is far away from the previous one, add a line break
@@ -751,7 +756,7 @@ def to_markdown(
             ):
                 if i in parms.written_tables:
                     continue
-                this_md += parms.tabs[i].to_markdown(clean=False)
+                this_md += parms.tabs[i].to_markdown(clean=False) + "\n"
                 if EXTRACT_WORDS:
                     # for "words" extraction, add table cells as line rects
                     cells = sorted(
@@ -772,7 +777,7 @@ def to_markdown(
             for i, trect in parms.tab_rects.items():
                 if i in parms.written_tables:
                     continue
-                this_md += parms.tabs[i].to_markdown(clean=False)
+                this_md += parms.tabs[i].to_markdown(clean=False) + "\n"
                 if EXTRACT_WORDS:
                     # for "words" extraction, add table cells as line rects
                     cells = sorted(
@@ -954,7 +959,7 @@ def to_markdown(
         )  # accept invisible text
 
         # determine background color
-        parms.bg_color = get_bg_color(page) if DETECT_BG_COLOR else None
+        parms.bg_color = None if not DETECT_BG_COLOR else get_bg_color(page)
 
         left, top, right, bottom = margins
         parms.clip = page.rect + (left, top, -right, -bottom)
@@ -1020,35 +1025,38 @@ def to_markdown(
         graphics_count = len([b for b in page.get_bboxlog() if "path" in b[0]])
         if GRAPHICS_LIMIT and graphics_count > GRAPHICS_LIMIT:
             IGNORE_GRAPHICS = True
+            too_many_graphics = True
+        else:
+            too_many_graphics = False
 
         # Locate all tables on page
         parms.written_tables = []  # stores already written tables
         omitted_table_rects = []
+        parms.tabs = []
         if IGNORE_GRAPHICS or not table_strategy:
             # do not try to extract tables
-            parms.tabs = None
+            pass
         else:
-            parms.tabs = page.find_tables(clip=parms.clip, strategy=table_strategy)
-            # remove tables with too few rows or columns
-            for i in range(len(parms.tabs.tables) - 1, -1, -1):
-                t = parms.tabs.tables[i]
+            tabs = page.find_tables(clip=parms.clip, strategy=table_strategy)
+            for t in tabs.tables:
+                # remove tables with too few rows or columns
                 if t.row_count < 2 or t.col_count < 2:
                     omitted_table_rects.append(pymupdf.Rect(t.bbox))
-                    del parms.tabs.tables[i]
-            parms.tabs.tables.sort(key=lambda t: (t.bbox[0], t.bbox[1]))
+                    continue
+                parms.tabs.append(t)
+            parms.tabs.sort(key=lambda t: (t.bbox[0], t.bbox[1]))
 
         # Make a list of table boundary boxes.
         # Must include the header bbox (which may exist outside tab.bbox)
         tab_rects = {}
-        if parms.tabs is not None:
-            for i, t in enumerate(parms.tabs.tables):
-                tab_rects[i] = pymupdf.Rect(t.bbox) | pymupdf.Rect(t.header.bbox)
-                tab_dict = {
-                    "bbox": tuple(tab_rects[i]),
-                    "rows": t.row_count,
-                    "columns": t.col_count,
-                }
-                parms.tables.append(tab_dict)
+        for i, t in enumerate(parms.tabs):
+            tab_rects[i] = pymupdf.Rect(t.bbox) | pymupdf.Rect(t.header.bbox)
+            tab_dict = {
+                "bbox": tuple(tab_rects[i]),
+                "rows": t.row_count,
+                "columns": t.col_count,
+            }
+            parms.tables.append(tab_dict)
         parms.tab_rects = tab_rects
         # list of table rectangles
         parms.tab_rects0 = list(tab_rects.values())
@@ -1064,17 +1072,14 @@ def to_markdown(
                 and p["rect"].width < parms.clip.width
                 and p["rect"].height < parms.clip.height
                 and (p["rect"].width > 3 or p["rect"].height > 3)
-                and not (p["fill"] == parms.bg_color and p["fill"] != None)
-                and not intersects_rects(
-                    p["rect"], parms.tab_rects0 + omitted_table_rects
-                )
+                and not (p["type"] == "f" and p["fill"] == parms.bg_color)
+                and not intersects_rects(p["rect"], parms.tab_rects0)
                 and not intersects_rects(p["rect"], parms.annot_rects)
             ]
         else:
             paths = []
-
         # catch too-many-graphics situation
-        if GRAPHICS_LIMIT and len(paths) > GRAPHICS_LIMIT:
+        if IGNORE_GRAPHICS:
             paths = []
 
         # We also ignore vector graphics that only represent
@@ -1100,17 +1105,27 @@ def to_markdown(
         parms.vg_clusters0 = refine_boxes(vg_clusters0)
 
         parms.vg_clusters = dict((i, r) for i, r in enumerate(parms.vg_clusters0))
+        block_count = len(parms.textpage.extractBLOCKS())
+        if block_count > 0:
+            char_density = len(parms.textpage.extractTEXT()) / block_count
+        else:
+            char_density = 0
         # identify text bboxes on page, avoiding tables, images and graphics
-        text_rects = column_boxes(
-            parms.page,
-            paths=parms.actual_paths,
-            no_image_text=not force_text,
-            textpage=parms.textpage,
-            avoid=parms.tab_rects0 + parms.vg_clusters0,
-            footer_margin=margins[3],
-            header_margin=margins[1],
-            ignore_images=IGNORE_IMAGES,
-        )
+        if too_many_graphics and char_density < 20:
+            # This page has too many isolated text pieces for meaningful
+            # layout analysis. Treat whole page as one text block.
+            text_rects = [parms.clip]
+        else:
+            text_rects = column_boxes(
+                parms.page,
+                paths=parms.actual_paths,
+                no_image_text=not force_text,
+                textpage=parms.textpage,
+                avoid=parms.tab_rects0 + parms.vg_clusters0,
+                footer_margin=margins[3],
+                header_margin=margins[1],
+                ignore_images=IGNORE_IMAGES,
+            )
 
         """
         ------------------------------------------------------------------
@@ -1168,6 +1183,9 @@ def to_markdown(
         else:
             words = []
         parms.words = words
+        if page_separators:
+            # add page separators to output
+            parms.md_string += f"\n\n--- end of page={parms.page.number} ---\n\n"
         return parms
 
     if page_chunks is False:
@@ -1196,7 +1214,13 @@ def to_markdown(
         pages = ProgressBar(pages)
     for pno in pages:
         parms = get_page_output(
-            doc, pno, margins, textflags, FILENAME, IGNORE_IMAGES, IGNORE_GRAPHICS
+            doc,
+            pno,
+            margins,
+            textflags,
+            FILENAME,
+            IGNORE_IMAGES,
+            IGNORE_GRAPHICS,
         )
         if page_chunks is False:
             document_output += parms.md_string
